@@ -5,20 +5,26 @@ import glob
 import io
 import logging
 import re
+from typing import Any
+from typing import Callable
+from typing import cast
 from typing import Dict
+from typing import Iterable
 from typing import List
+from typing import Mapping
 from typing import Optional
+from typing import Sequence
 from typing import Set
 from typing import Tuple
 from typing import Union
 
+from certbot_nginx._internal import nginxparser
+from certbot_nginx._internal import obj
+from certbot_nginx._internal.nginxparser import UnspacedList
 import pyparsing
 
 from certbot import errors
 from certbot.compat import os
-from certbot_nginx._internal import nginxparser
-from certbot_nginx._internal import obj
-from certbot_nginx._internal.nginxparser import UnspacedList
 
 logger = logging.getLogger(__name__)
 
@@ -32,8 +38,8 @@ class NginxParser:
 
     """
 
-    def __init__(self, root):
-        self.parsed: Dict[str, Union[List, nginxparser.UnspacedList]] = {}
+    def __init__(self, root: str) -> None:
+        self.parsed: Dict[str, UnspacedList] = {}
         self.root = os.path.abspath(root)
         self.config_root = self._find_config_root()
 
@@ -42,14 +48,14 @@ class NginxParser:
         # not enable sites from there.
         self.load()
 
-    def load(self):
+    def load(self) -> None:
         """Loads Nginx files into a parsed tree.
 
         """
         self.parsed = {}
         self._parse_recursively(self.config_root)
 
-    def _parse_recursively(self, filepath):
+    def _parse_recursively(self, filepath: str) -> None:
         """Parses nginx config files recursively by looking at 'include'
         directives inside 'http' and 'server' blocks. Note that this only
         reads Nginx files that potentially declare a virtual host.
@@ -77,7 +83,7 @@ class NginxParser:
                                 if _is_include_directive(server_entry):
                                     self._parse_recursively(server_entry[1])
 
-    def abs_path(self, path):
+    def abs_path(self, path: str) -> str:
         """Converts a relative path to an absolute path relative to the root.
         Does nothing for paths that are already absolute.
 
@@ -90,14 +96,14 @@ class NginxParser:
             return os.path.normpath(os.path.join(self.root, path))
         return os.path.normpath(path)
 
-    def _build_addr_to_ssl(self):
+    def _build_addr_to_ssl(self) -> Dict[Tuple[str, str], bool]:
         """Builds a map from address to whether it listens on ssl in any server block
         """
         servers = self._get_raw_servers()
 
         addr_to_ssl: Dict[Tuple[str, str], bool] = {}
-        for filename in servers:
-            for server, _ in servers[filename]:
+        for server_list in servers.values():
+            for server, _ in server_list:
                 # Parse the server block to save addr info
                 parsed_server = _parse_server_raw(server)
                 for addr in parsed_server['addrs']:
@@ -107,13 +113,12 @@ class NginxParser:
                     addr_to_ssl[addr_tuple] = addr.ssl or addr_to_ssl[addr_tuple]
         return addr_to_ssl
 
-    def _get_raw_servers(self) -> Dict:
+    def _get_raw_servers(self) -> Dict[str, Union[List[Any], UnspacedList]]:
         # pylint: disable=cell-var-from-loop
         """Get a map of unparsed all server blocks
         """
-        servers: Dict[str, Union[List, nginxparser.UnspacedList]] = {}
-        for filename in self.parsed:
-            tree = self.parsed[filename]
+        servers: Dict[str, Union[List[Any], nginxparser.UnspacedList]] = {}
+        for filename, tree in self.parsed.items():
             servers[filename] = []
             srv = servers[filename]  # workaround undefined loop var in lambdas
 
@@ -127,7 +132,7 @@ class NginxParser:
                 servers[filename][i] = (new_server, path)
         return servers
 
-    def get_vhosts(self):
+    def get_vhosts(self) -> List[obj.VirtualHost]:
         """Gets list of all 'virtual hosts' found in Nginx configuration.
         Technically this is a misnomer because Nginx does not have virtual
         hosts, it has 'server blocks'.
@@ -141,8 +146,8 @@ class NginxParser:
         servers = self._get_raw_servers()
 
         vhosts = []
-        for filename in servers:
-            for server, path in servers[filename]:
+        for filename, server_list in servers.items():
+            for server, path in server_list:
                 # Parse the server block into a VirtualHost object
 
                 parsed_server = _parse_server_raw(server)
@@ -159,7 +164,7 @@ class NginxParser:
 
         return vhosts
 
-    def _update_vhosts_addrs_ssl(self, vhosts):
+    def _update_vhosts_addrs_ssl(self, vhosts: Iterable[obj.VirtualHost]) -> None:
         """Update a list of raw parsed vhosts to include global address sslishness
         """
         addr_to_ssl = self._build_addr_to_ssl()
@@ -169,7 +174,7 @@ class NginxParser:
                 if addr.ssl:
                     vhost.ssl = True
 
-    def _get_included_directives(self, block):
+    def _get_included_directives(self, block: UnspacedList) -> UnspacedList:
         """Returns array with the "include" directives expanded out by
         concatenating the contents of the included file to the block.
 
@@ -189,7 +194,7 @@ class NginxParser:
                         pass
         return result
 
-    def _parse_files(self, filepath, override=False):
+    def _parse_files(self, filepath: str, override: bool = False) -> List[UnspacedList]:
         """Parse files from a glob
 
         :param str filepath: Nginx config file path
@@ -220,7 +225,7 @@ class NginxParser:
                 logger.warning("Could not parse file: %s due to %s", item, err)
         return trees
 
-    def _find_config_root(self):
+    def _find_config_root(self) -> str:
         """Return the Nginx Configuration Root file."""
         location = ['nginx.conf']
 
@@ -231,7 +236,7 @@ class NginxParser:
         raise errors.NoInstallationError(
             "Could not find Nginx root configuration file (nginx.conf)")
 
-    def filedump(self, ext='tmp', lazy=True):
+    def filedump(self, ext: str = 'tmp', lazy: bool = True) -> None:
         """Dumps parsed configurations into files.
 
         :param str ext: The file extension to use for the dumped files. If
@@ -240,8 +245,7 @@ class NginxParser:
 
         """
         # Best-effort atomicity is enforced above us by reverter.py
-        for filename in self.parsed:
-            tree = self.parsed[filename]
+        for filename, tree in self.parsed.items():
             if ext:
                 filename = filename + os.path.extsep + ext
             if not isinstance(tree, UnspacedList):
@@ -257,7 +261,7 @@ class NginxParser:
             except IOError:
                 logger.error("Could not open file for writing: %s", filename)
 
-    def parse_server(self, server):
+    def parse_server(self, server: UnspacedList) -> Dict[str, Any]:
         """Parses a list of server directives, accounting for global address sslishness.
 
         :param list server: list of directives in a server block
@@ -268,7 +272,7 @@ class NginxParser:
         _apply_global_addr_ssl(addr_to_ssl, parsed_server)
         return parsed_server
 
-    def has_ssl_on_directive(self, vhost):
+    def has_ssl_on_directive(self, vhost: obj.VirtualHost) -> bool:
         """Does vhost have ssl on for all ports?
 
         :param :class:`~certbot_nginx._internal.obj.VirtualHost` vhost: The vhost in question
@@ -286,7 +290,8 @@ class NginxParser:
 
         return False
 
-    def add_server_directives(self, vhost, directives, insert_at_top=False):
+    def add_server_directives(self, vhost: obj.VirtualHost, directives: List[Any],
+                              insert_at_top: bool = False) -> None:
         """Add directives to the server block identified by vhost.
 
         This method modifies vhost to be fully consistent with the new directives.
@@ -307,7 +312,8 @@ class NginxParser:
         self._modify_server_directives(vhost,
             functools.partial(_add_directives, directives, insert_at_top))
 
-    def update_or_add_server_directives(self, vhost, directives, insert_at_top=False):
+    def update_or_add_server_directives(self, vhost: obj.VirtualHost, directives: List[Any],
+                                        insert_at_top: bool = False) -> None:
         """Add or replace directives in the server block identified by vhost.
 
         This method modifies vhost to be fully consistent with the new directives.
@@ -329,7 +335,8 @@ class NginxParser:
         self._modify_server_directives(vhost,
             functools.partial(_update_or_add_directives, directives, insert_at_top))
 
-    def remove_server_directives(self, vhost, directive_name, match_func=None):
+    def remove_server_directives(self, vhost: obj.VirtualHost, directive_name: str,
+                                 match_func: Optional[Callable[[Any], bool]] = None) -> None:
         """Remove all directives of type directive_name.
 
         :param :class:`~certbot_nginx._internal.obj.VirtualHost` vhost: The vhost
@@ -341,7 +348,8 @@ class NginxParser:
         self._modify_server_directives(vhost,
             functools.partial(_remove_directives, directive_name, match_func))
 
-    def _update_vhost_based_on_new_directives(self, vhost, directives_list):
+    def _update_vhost_based_on_new_directives(self, vhost: obj.VirtualHost,
+                                              directives_list: UnspacedList) -> None:
         new_server = self._get_included_directives(directives_list)
         parsed_server = self.parse_server(new_server)
         vhost.addrs = parsed_server['addrs']
@@ -349,7 +357,8 @@ class NginxParser:
         vhost.names = parsed_server['names']
         vhost.raw = new_server
 
-    def _modify_server_directives(self, vhost, block_func):
+    def _modify_server_directives(self, vhost: obj.VirtualHost,
+                                  block_func: Callable[[List[Any]], None]) -> None:
         filename = vhost.filep
         try:
             result = self.parsed[filename]
@@ -366,7 +375,7 @@ class NginxParser:
 
     def duplicate_vhost(self, vhost_template: obj.VirtualHost,
                         remove_singleton_listen_params: bool = False,
-                        only_directives: Optional[List] = None) -> obj.VirtualHost:
+                        only_directives: Optional[List[Any]] = None) -> obj.VirtualHost:
         """Duplicate the vhost in the configuration files.
 
         :param :class:`~certbot_nginx._internal.obj.VirtualHost` vhost_template: The vhost
@@ -419,7 +428,7 @@ class NginxParser:
         return new_vhost
 
 
-def _parse_ssl_options(ssl_options):
+def _parse_ssl_options(ssl_options: Optional[str]) -> List[UnspacedList]:
     if ssl_options is not None:
         try:
             with io.open(ssl_options, "r", encoding="utf-8") as _file:
@@ -431,9 +440,12 @@ def _parse_ssl_options(ssl_options):
                            "Only UTF-8 encoding is supported.", ssl_options)
         except pyparsing.ParseBaseException as err:
             logger.warning("Could not parse file: %s due to %s", ssl_options, err)
-    return []
+    return UnspacedList([])
 
-def _do_for_subarray(entry, condition, func, path=None):
+
+def _do_for_subarray(entry: List[Any], condition: Callable[[List[Any]], bool],
+                     func: Callable[[List[Any], List[int]], None],
+                     path: Optional[List[int]] = None) -> None:
     """Executes a function for a subarray of a nested array if it matches
     the given condition.
 
@@ -452,7 +464,7 @@ def _do_for_subarray(entry, condition, func, path=None):
                 _do_for_subarray(item, condition, func, path + [index])
 
 
-def get_best_match(target_name, names):
+def get_best_match(target_name: str, names: Iterable[str]) -> Tuple[Optional[str], Optional[str]]:
     """Finds the best match for target_name out of names using the Nginx
     name-matching rules (exact > longest wildcard starting with * >
     longest wildcard ending with * > regex).
@@ -481,29 +493,29 @@ def get_best_match(target_name, names):
     if exact:
         # There can be more than one exact match; e.g. eff.org, .eff.org
         match = min(exact, key=len)
-        return ('exact', match)
+        return 'exact', match
     if wildcard_start:
         # Return the longest wildcard
         match = max(wildcard_start, key=len)
-        return ('wildcard_start', match)
+        return 'wildcard_start', match
     if wildcard_end:
         # Return the longest wildcard
         match = max(wildcard_end, key=len)
-        return ('wildcard_end', match)
+        return 'wildcard_end', match
     if regex:
         # Just return the first one for now
         match = regex[0]
-        return ('regex', match)
+        return 'regex', match
 
-    return (None, None)
+    return None, None
 
 
-def _exact_match(target_name, name):
+def _exact_match(target_name: str, name: str) -> bool:
     target_lower = target_name.lower()
     return name.lower() in (target_lower, '.' + target_lower)
 
 
-def _wildcard_match(target_name, name, start):
+def _wildcard_match(target_name: str, name: str, start: bool) -> bool:
     # Degenerate case
     if name == '*':
         return True
@@ -528,7 +540,7 @@ def _wildcard_match(target_name, name, start):
     return target_name_lower.endswith('.' + name_lower)
 
 
-def _regex_match(target_name, name):
+def _regex_match(target_name: str, name: str) -> bool:
     # Must start with a tilde
     if len(name) < 2 or name[0] != '~':
         return False
@@ -536,13 +548,13 @@ def _regex_match(target_name, name):
     # After tilde is a perl-compatible regex
     try:
         regex = re.compile(name[1:])
-        return re.match(regex, target_name)
+        return bool(re.match(regex, target_name))
     except re.error:  # pragma: no cover
         # perl-compatible regexes are sometimes not recognized by python
         return False
 
 
-def _is_include_directive(entry):
+def _is_include_directive(entry: Any) -> bool:
     """Checks if an nginx parsed entry is an 'include' directive.
 
     :param list entry: the parsed entry
@@ -554,7 +566,8 @@ def _is_include_directive(entry):
             len(entry) == 2 and entry[0] == 'include' and
             isinstance(entry[1], str))
 
-def _is_ssl_on_directive(entry):
+
+def _is_ssl_on_directive(entry: Any) -> bool:
     """Checks if an nginx parsed entry is an 'ssl on' directive.
 
     :param list entry: the parsed entry
@@ -566,14 +579,18 @@ def _is_ssl_on_directive(entry):
             len(entry) == 2 and entry[0] == 'ssl' and
             entry[1] == 'on')
 
-def _add_directives(directives, insert_at_top, block):
+
+def _add_directives(directives: List[Any], insert_at_top: bool,
+                    block: UnspacedList) -> None:
     """Adds directives to a config block."""
     for directive in directives:
         _add_directive(block, directive, insert_at_top)
     if block and '\n' not in block[-1]:  # could be "   \n  " or ["\n"] !
         block.append(nginxparser.UnspacedList('\n'))
 
-def _update_or_add_directives(directives, insert_at_top, block):
+
+def _update_or_add_directives(directives: List[Any], insert_at_top: bool,
+                              block: UnspacedList) -> None:
     """Adds or replaces directives in a config block."""
     for directive in directives:
         _update_or_add_directive(block, directive, insert_at_top)
@@ -586,7 +603,8 @@ REPEATABLE_DIRECTIVES = {'server_name', 'listen', INCLUDE, 'rewrite', 'add_heade
 COMMENT = ' managed by Certbot'
 COMMENT_BLOCK = [' ', '#', COMMENT]
 
-def comment_directive(block, location):
+
+def comment_directive(block: UnspacedList, location: int) -> None:
     """Add a ``#managed by Certbot`` comment to the end of the line at location.
 
     :param list block: The block containing the directive to be commented
@@ -605,40 +623,45 @@ def comment_directive(block, location):
     if next_entry is not None and "\n" not in next_entry:
         block.insert(location + 2, '\n')
 
-def _comment_out_directive(block, location, include_location):
+
+def _comment_out_directive(block: UnspacedList, location: int, include_location: str) -> None:
     """Comment out the line at location, with a note of explanation."""
     comment_message = ' duplicated in {0}'.format(include_location)
     # add the end comment
     # create a dumpable object out of block[location] (so it includes the ;)
     directive = block[location]
-    new_dir_block = nginxparser.UnspacedList([]) # just a wrapper
+    new_dir_block = nginxparser.UnspacedList([])  # just a wrapper
     new_dir_block.append(directive)
     dumped = nginxparser.dumps(new_dir_block)
-    commented = dumped + ' #' + comment_message # add the comment directly to the one-line string
-    new_dir = nginxparser.loads(commented) # reload into UnspacedList
+    commented = dumped + ' #' + comment_message  # add the comment directly to the one-line string
+    new_dir = nginxparser.loads(commented)  # reload into UnspacedList
 
     # add the beginning comment
     insert_location = 0
-    if new_dir[0].spaced[0] != new_dir[0][0]: # if there's whitespace at the beginning
+    if new_dir[0].spaced[0] != new_dir[0][0]:  # if there's whitespace at the beginning
         insert_location = 1
-    new_dir[0].spaced.insert(insert_location, "# ") # comment out the line
-    new_dir[0].spaced.append(";") # directly add in the ;, because now dumping won't work properly
+    new_dir[0].spaced.insert(insert_location, "# ")  # comment out the line
+    new_dir[0].spaced.append(";")  # directly add in the ;, because now dumping won't work properly
     dumped = nginxparser.dumps(new_dir)
-    new_dir = nginxparser.loads(dumped) # reload into an UnspacedList
+    new_dir = nginxparser.loads(dumped)  # reload into an UnspacedList
 
     block[location] = new_dir[0] # set the now-single-line-comment directive back in place
 
-def _find_location(block, directive_name, match_func=None):
+
+def _find_location(block: UnspacedList, directive_name: str,
+                   match_func: Optional[Callable[[Any], bool]] = None) -> Optional[int]:
     """Finds the index of the first instance of directive_name in block.
        If no line exists, use None."""
-    return next((index for index, line in enumerate(block) \
-        if line and line[0] == directive_name and (match_func is None or match_func(line))), None)
+    return next((index for index, line in enumerate(block) if (
+        line and line[0] == directive_name and (match_func is None or match_func(line)))), None)
 
-def _is_whitespace_or_comment(directive):
+
+def _is_whitespace_or_comment(directive: Sequence[Any]) -> bool:
     """Is this directive either a whitespace or comment directive?"""
     return len(directive) == 0 or directive[0] == '#'
 
-def _add_directive(block, directive, insert_at_top):
+
+def _add_directive(block: UnspacedList, directive: Sequence[Any], insert_at_top: bool) -> None:
     if not isinstance(directive, nginxparser.UnspacedList):
         directive = nginxparser.UnspacedList(directive)
     if _is_whitespace_or_comment(directive):
@@ -655,10 +678,11 @@ def _add_directive(block, directive, insert_at_top):
     # handle flat include files
 
     directive_name = directive[0]
-    def can_append(loc, dir_name):
+
+    def can_append(loc: Optional[int], dir_name: str) -> bool:
         """ Can we append this directive to the block? """
         return loc is None or (isinstance(dir_name, str)
-            and dir_name in REPEATABLE_DIRECTIVES)
+                               and dir_name in REPEATABLE_DIRECTIVES)
 
     err_fmt = 'tried to insert directive "{0}" but found conflicting "{1}".'
 
@@ -674,10 +698,14 @@ def _add_directive(block, directive, insert_at_top):
             included_dir_name = included_directive[0]
             if (not _is_whitespace_or_comment(included_directive)
                     and not can_append(included_dir_loc, included_dir_name)):
-                if block[included_dir_loc] != included_directive:
-                    raise errors.MisconfigurationError(err_fmt.format(included_directive,
-                        block[included_dir_loc]))
-                _comment_out_directive(block, included_dir_loc, directive[1])
+
+                # By construction of can_append(), included_dir_loc cannot be None at that point
+                resolved_included_dir_loc = cast(int, included_dir_loc)
+
+                if block[resolved_included_dir_loc] != included_directive:
+                    raise errors.MisconfigurationError(err_fmt.format(
+                        included_directive, block[resolved_included_dir_loc]))
+                _comment_out_directive(block, resolved_included_dir_loc, directive[1])
 
     if can_append(location, directive_name):
         if insert_at_top:
@@ -689,14 +717,22 @@ def _add_directive(block, directive, insert_at_top):
         else:
             block.append(directive)
             comment_directive(block, len(block) - 1)
-    elif block[location] != directive:
-        raise errors.MisconfigurationError(err_fmt.format(directive, block[location]))
+        return
 
-def _update_directive(block, directive, location):
+    # By construction of can_append(), location cannot be None at that point
+    resolved_location = cast(int, location)
+
+    if block[resolved_location] != directive:
+        raise errors.MisconfigurationError(err_fmt.format(directive, block[resolved_location]))
+
+
+def _update_directive(block: UnspacedList, directive: Sequence[Any], location: int) -> None:
     block[location] = directive
     comment_directive(block, location)
 
-def _update_or_add_directive(block, directive, insert_at_top):
+
+def _update_or_add_directive(block: UnspacedList, directive: Sequence[Any],
+                             insert_at_top: bool) -> None:
     if not isinstance(directive, nginxparser.UnspacedList):
         directive = nginxparser.UnspacedList(directive)
     if _is_whitespace_or_comment(directive):
@@ -713,10 +749,13 @@ def _update_or_add_directive(block, directive, insert_at_top):
 
     _add_directive(block, directive, insert_at_top)
 
-def _is_certbot_comment(directive):
+
+def _is_certbot_comment(directive: Sequence[Any]) -> bool:
     return '#' in directive and COMMENT in directive
 
-def _remove_directives(directive_name, match_func, block):
+
+def _remove_directives(directive_name: str, match_func: Callable[[Any], bool],
+                       block: UnspacedList) -> None:
     """Removes directives of name directive_name from a config block if match_func matches.
     """
     while True:
@@ -728,7 +767,9 @@ def _remove_directives(directive_name, match_func, block):
             del block[location + 1]
         del block[location]
 
-def _apply_global_addr_ssl(addr_to_ssl, parsed_server):
+
+def _apply_global_addr_ssl(addr_to_ssl: Mapping[Tuple[str, str], bool],
+                           parsed_server: Dict[str, Any]) -> None:
     """Apply global sslishness information to the parsed server block
     """
     for addr in parsed_server['addrs']:
@@ -736,7 +777,8 @@ def _apply_global_addr_ssl(addr_to_ssl, parsed_server):
         if addr.ssl:
             parsed_server['ssl'] = True
 
-def _parse_server_raw(server):
+
+def _parse_server_raw(server: UnspacedList) -> Dict[str, Any]:
     """Parses a list of server directives.
 
     :param list server: list of directives in a server block
